import fc from 'fast-check';
import { beforeEach, describe, expect, it, vi } from 'vitest';
import type { DID } from '../../types';
import { ConnectionManager, TransactionType } from '../connections';
import { seal } from '../encryption';

// Mock fetch globally
const fetchMock = vi.fn();
global.fetch = fetchMock;

// Helper to generate valid DID strings
const arbDID = fc
  .tuple(fc.string(), fc.string())
  .map(([method, id]) => `did:${method}:${id}` as DID);

// Test helper to create arbitrary ComponentRequests
const arbComponentRequest = fc.record({
  tx: fc.record({
    requestId: fc.option(fc.uuid(), { nil: undefined }),
    customer: fc.option(
      fc.record({
        name: fc.string(),
        email: fc.option(fc.string(), { nil: undefined }),
        phone: fc.option(fc.string(), { nil: undefined }),
        type: fc.constant(undefined), // Optional PersonType
        accountNumber: fc.option(fc.string(), { nil: undefined }),
        did: fc.option(arbDID, { nil: undefined }),
        verified: fc.option(fc.boolean(), { nil: undefined }),
        website: fc.option(fc.webUrl(), { nil: undefined }),
        geographicAddress: fc.option(fc.constant(undefined), {
          nil: undefined,
        }),
        nationalIdentification: fc.option(fc.constant(undefined), {
          nil: undefined,
        }),
      }),
      { nil: undefined },
    ),
  }),
  authToken: fc.option(fc.string(), { nil: undefined }),
  txOptions: fc.option(fc.record({}), { nil: undefined }),
});

// New arbitrary for ConnectionMetadata
const arbConnectionMetadata = fc.record({
  participants: fc.array(fc.string(), { minLength: 1 }), // At least one participant
  nodeUrl: fc.webUrl(),
  transactionType: fc.constantFrom<TransactionType>('withdraw', 'deposit'),
});

describe('ConnectionManager', () => {
  let manager: ConnectionManager;
  const testEndpoint = 'https://test-endpoint.com';

  beforeEach(() => {
    manager = new ConnectionManager(testEndpoint);
    fetchMock.mockReset();
  });

  describe('create', () => {
    it('should successfully create a new connection', async () => {
      await fc.assert(
        fc.asyncProperty(
          arbComponentRequest,
          arbConnectionMetadata,
          async (request, metadata) => {
            // Mock successful response
            const mockResponse = {
              id: 'test-id',
              version: 1,
              metadata,
              sealed: ['encrypted-data'],
            };

            fetchMock.mockResolvedValueOnce({
              ok: true,
              json: async () => mockResponse,
            });

            const result = await manager.create(request, metadata);

            // Verify fetch was called correctly
            expect(fetchMock).toHaveBeenCalledWith(testEndpoint, {
              method: 'POST',
              headers: {
                'Content-Type': 'application/json',
              },
              body: expect.any(String),
            });

            // Verify response structure
            expect(result).toEqual({
              id: mockResponse.id,
              metadata: mockResponse.metadata,
              version: mockResponse.version,
              data: request,
              key: expect.any(String),
            });
          },
        ),
      );
    });

    it('should throw error on failed creation', async () => {
      await fc.assert(
        fc.asyncProperty(
          arbComponentRequest,
          arbConnectionMetadata,
          async (request, metadata) => {
            // Mock failed response
            fetchMock.mockResolvedValueOnce({
              ok: false,
              text: async () => 'Creation failed',
            });

            await expect(manager.create(request, metadata)).rejects.toThrow(
              'Failed to create connection',
            );
          },
        ),
      );
    });
  });

  describe('update', () => {
    it('should successfully update an existing connection', async () => {
      await fc.assert(
        fc.asyncProperty(
          fc.uuid(),
          arbComponentRequest,
          fc.integer(),
          async (id, request, version) => {
            // Mock successful response
            const mockResponse = {
              id,
              metadata: arbConnectionMetadata,
              version: version + 1,
              sealed: ['new-encrypted-data'],
            };

            fetchMock.mockResolvedValueOnce({
              ok: true,
              json: async () => mockResponse,
            });

            const result = await manager.update(id, request, version);

            // Verify fetch was called correctly
            expect(fetchMock).toHaveBeenCalledWith(`${testEndpoint}/${id}`, {
              method: 'PATCH',
              headers: {
                'Content-Type': 'application/json',
              },
              body: expect.any(String),
            });

            // Verify response structure
            expect(result).toEqual({
              id: mockResponse.id,
              version: mockResponse.version,
              metadata: mockResponse.metadata,
              data: request,
              key: expect.any(String),
            });
          },
        ),
      );
    });

    it('should throw error on failed update', async () => {
      await fc.assert(
        fc.asyncProperty(
          fc.uuid(),
          arbComponentRequest,
          fc.integer(),
          async (id, request, version) => {
            // Mock failed response
            fetchMock.mockResolvedValueOnce({
              ok: false,
              text: async () => 'Update failed',
            });

            await expect(manager.update(id, request, version)).rejects.toThrow(
              'Failed to update connection',
            );
          },
        ),
      );
    });
  });

  describe('get', () => {
    it('should successfully retrieve and decrypt connection data', async () => {
      // Create a real sealed object to use in our test
      const testData = { requestId: 'test-123' };
      const sealed = await seal(testData);

      await fc.assert(
        fc.asyncProperty(fc.uuid(), async (id) => {
          // Mock successful response with real encrypted data
          const mockResponse = {
            id,
            version: 1,
            metadata: arbConnectionMetadata,
            sealed: [sealed.ciphertext], // Use the real ciphertext
          };

          fetchMock.mockResolvedValueOnce({
            ok: true,
            json: async () => mockResponse,
          });

          const result = await manager.get(id, sealed.key); // Use the matching key

          // Verify fetch was called correctly
          expect(fetchMock).toHaveBeenCalledWith(`${testEndpoint}/${id}`, {
            method: 'GET',
          });

          // Verify response structure
          expect(result).toEqual({
            id: mockResponse.id,
            version: mockResponse.version,
            metadata: mockResponse.metadata,
            data: testData, // Should match our original test data
            key: sealed.key,
          });
        }),
      );
    });

    it('should throw error on failed retrieval', async () => {
      await fc.assert(
        fc.asyncProperty(fc.uuid(), fc.string(), async (id, key) => {
          // Mock failed response
          fetchMock.mockResolvedValueOnce({
            ok: false,
            text: async () => 'Retrieval failed',
          });

          await expect(manager.get(id, key)).rejects.toThrow(
            'Failed to get connection',
          );
        }),
      );
    });
  });

  describe('close', () => {
    it('should successfully close a connection', async () => {
      await fc.assert(
        fc.asyncProperty(fc.uuid(), async (id) => {
          // Mock successful response
          fetchMock.mockResolvedValueOnce({
            ok: true,
          });

          await manager.close(id);

          // Verify fetch was called correctly
          expect(fetchMock).toHaveBeenCalledWith(`${testEndpoint}/${id}`, {
            method: 'DELETE',
          });
        }),
      );
    });

    it('should throw error on failed closure', async () => {
      await fc.assert(
        fc.asyncProperty(fc.uuid(), async (id) => {
          // Mock failed response
          fetchMock.mockResolvedValueOnce({
            ok: false,
            text: async () => 'Closure failed',
          });

          await expect(manager.close(id)).rejects.toThrow(
            'Failed to close connection',
          );
        }),
      );
    });
  });
});
